from discoruns.mechanism import Mechanism
from discoruns.wrapper.fs_wrapper import ForensicStoreWrapper


def _process_artifacts(tmp_dict, fsw, artifact_name, value_name):
    for artifact in fsw.get_artifacts(artifact_name):
        for value in artifact.get("values", []):
            if value.get("data") and value.get("name").casefold() == value_name.casefold():
                tmp_dict.setdefault(value.get("data"), set()).add(artifact.get("key"))


class WinlogonHelper(Mechanism):

    def collect_mechanism(self, fsw: ForensicStoreWrapper) -> list:

        tmp_dict = {}
        artifact_data = [("WindowsWinlogonGinaDLL", "GinaDLL"),
                         ("WindowsWinlogonNotify", "DLLName"),
                         ("WindowsWinlogonShell", "Shell"),
                         ("WindowsWinlogonSystem", "System"),
                         ("WindowsWinlogonTaskman", "Taskman"),
                         ("WindowsWinlogonUiHost", "UiHost"),
                         ("WindowsWinlogonUserinit", "Userinit"),
                         ("WindowsWinlogonVMApplet", "VMApplet"),
                         ("WindowsWinlogonAppSetup", "AppSetup"),
                         ("WindowsWinlogonGPExtensions", "DllName")]

        for artifact_name, value_name in artifact_data:
            _process_artifacts(tmp_dict, fsw, artifact_name, value_name)

        return [{"name": "winlogon_helper", "origins": list(origins), "entry": helper}
                for helper, origins in tmp_dict.items()]



import logging
import sqlite3

import forensicstore

logger = logging.getLogger(__name__)


class ForensicStoreWrapper:
    """
    This is a wrapper class for the forensicstore database.
    """

    def __init__(self, forensicstore_path: str):
        """
        Initializes forensicstore wrapper class.

        :param forensicstore_path: Path to an existing forensicstore.
        """

        self.forensicstore_path = forensicstore_path
        self._fs_store = forensicstore.open(self.forensicstore_path)

    def get_artifacts(self, artifact_name: str):
        try:
            registry_values = self._fs_store.select([{"type": "windows-registry-key", "artifact": artifact_name}])

            logger.debug("Collecting '%s' persistence mechanism.", artifact_name)
            return list(registry_values)

        except sqlite3.OperationalError:
            return []

        except Exception as exception:
            logger.exception(exception)

    def get_files(self, artifact_name: str):
        try:
            registry_values = self._fs_store.select([{"type": "file", "artifact": artifact_name}])
            logger.debug("Collecting '%s' file information.", artifact_name)
            return list(registry_values)

        except sqlite3.OperationalError:
            return []

        except Exception as exception:
            logger.exception(exception)

    def close(self):
        self._fs_store.close()



from discoruns.mechanism import Mechanism
from discoruns.wrapper.fs_wrapper import ForensicStoreWrapper


class Services(Mechanism):

    def collect_mechanism(self, fsw: ForensicStoreWrapper) -> list:
        tmp_dict = {}
        mechanisms = []

        for artifact in fsw.get_artifacts("WindowsServices"):
            for value in artifact.get("values", []):
                name = value.get("name", "").casefold()
                if name == "ServiceDll".casefold() or name == "ImagePath".casefold() and value.get("data"):
                    tmp_dict.setdefault(value.get("data"), set()).add(artifact.get("key"))

        for artifact in fsw.get_artifacts("WindowsFontDriversAlt"):
            for value in artifact.get("values", []):
                tmp_dict.setdefault(value.get("data"), set()).add(artifact.get("key"))

        for entry, origins in tmp_dict.items():
            if entry.endswith(".sys"):
                mechanisms.append({"name": "driver", "origins": list(origins), "entry": entry})
            else:
                mechanisms.append({"name": "service", "origins": list(origins), "entry": entry})

        return mechanisms



import json
import os

from discoruns.mechanism import Mechanism
from discoruns.wrapper.fs_wrapper import ForensicStoreWrapper


class BrowserExtensions(Mechanism):

    def collect_mechanism(self, fsw: ForensicStoreWrapper) -> list:
        tmp_dict = {}
        folder_path = os.path.join(fsw.forensicstore_path, "WindowsBrowserPersistenceFiles/")
        if os.path.exists(folder_path):
            for file in os.listdir(folder_path):
                if file.endswith(".json"):
                    file_path = os.path.join(folder_path, file)
                    with open(file_path) as io:
                        for addon in json.load(io).get("addons"):
                            name = addon.get("defaultLocale", {}).get("name")
                            path = addon.get("path", "Unkonwn")
                            if addon.get("active") and name != "Default" and path:
                                tmp_dict.setdefault(name, set()).add(path)

        for artifact in fsw.get_artifacts("WindowsBrowserPersistenceKeys"):
            for value in artifact.get("values", []):
                if value.get("name") in ["(Default)", "ToolTip"]:
                    tmp_dict.setdefault(value.get("data"), set()).add(artifact.get("key"))

        return [{"name": "browser_extensions", "origins": list(origins), "entry": extension_data}
                for extension_data, origins in tmp_dict.items()]



from abc import ABC, abstractmethod
from discoruns.wrapper.fs_wrapper import ForensicStoreWrapper


class Mechanism(ABC):

    @abstractmethod
    def collect_mechanism(self, fsw: ForensicStoreWrapper) -> list:
        raise NotImplementedError



from discoruns.mechanism import Mechanism


class PowershellProfiles(Mechanism):

    def collect_mechanism(self, fsw):
        tmp_dict = {}

        for entry in fsw.get_files("WindowsPowerShellDefaultProfiles"):
            tmp_dict.setdefault(entry.get("origin").get("path"), []).extend(entry.get("name"))

        return [{"name": "powershell_profiles", "origins": [path],
                 "entry": entry.pop()} for path, entry in tmp_dict.items()]



from discoruns.mechanism import Mechanism
from discoruns.wrapper.fs_wrapper import ForensicStoreWrapper


def _process_artifacts(tmp_dict, fsw, artifact_name, value_name):
    for artifact in fsw.get_artifacts(artifact_name):
        for value in artifact.get("values", []):
            if value.get("data") and value.get("name").casefold() == value_name.casefold():
                tmp_dict.setdefault(value.get("data"), set()).add(artifact.get("key"))


class WinsockProviders(Mechanism):

    def collect_mechanism(self, fsw: ForensicStoreWrapper) -> list:
        tmp_dict = {}
        artifact_data = [("WinSock2LayeredServiceProviders", "LibraryPath"),
                         ("WinSock2NamespaceProviders", "LibraryPath")]

        for artifact_name, value_name in artifact_data:
            _process_artifacts(tmp_dict, fsw, artifact_name, value_name)

        return [{"name": "winsock_providers", "origins": list(origins), "entry": helper}
                for helper, origins in tmp_dict.items()]



from discoruns.mechanism import Mechanism
from discoruns.wrapper.fs_wrapper import ForensicStoreWrapper


class FileAssociations(Mechanism):

    def collect_mechanism(self, fsw: ForensicStoreWrapper) -> list:
        tmp_dict = {}

        for artifact in fsw.get_artifacts("WindowsFileAssociation"):
            if values := artifact.get("values"):

                # Find the last program the extension was opened with via MRUList.
                mru_lists = [value.get("data") for value in values if value.get("name") == "MRUList"]
                mru_list = ""
                if len(mru_lists) == 1:
                    mru_list = mru_lists[0]

                last_prog_char = ""
                if len(mru_list) >= 1:
                    last_prog_char = mru_list[0]

                for value in artifact.get("values", []):
                    if value.get("data") and value.get("name") == last_prog_char:
                        tmp_dict.setdefault(value.get("data"), set()).add(artifact.get("key"))

        return [{"name": "file_associations", "origins": list(origins), "entry": association}
                for association, origins in tmp_dict.items()]



import configparser
import os

from discoruns.mechanism import Mechanism
from discoruns.wrapper.fs_wrapper import ForensicStoreWrapper


def process_ini_values(tmp_dict: dict, file_path: str, section: str, file_origin: str):
    config = configparser.ConfigParser()
    config.read(file_path, encoding="UTF-16")

    logon_entries = zip(list(config[section])[::2], list(config[section])[1::2])
    for cmd_key, arg_key in logon_entries:
        tmp_dict.setdefault(config[section][cmd_key] + " " + config[section][arg_key], set()).add(file_origin)


def get_file_origin(fsw, file_path):
    return [file.get("origin").get("path") for file in fsw.get_files("WindowsGroupPolicyScripts")
            if file.get("export_path") in file_path]


class Logon(Mechanism):

    def collect_mechanism(self, fsw: ForensicStoreWrapper) -> list:
        mechanisms = []
        tmp_dict = {}

        # Windows Environment Logon Scripts
        for artifact in fsw.get_artifacts("WindowsEnvironmentUserLoginScripts"):
            for value in artifact.get("values", []):
                if value.get("data"):
                    tmp_dict.setdefault(value.get("data"), set()).add(artifact.get("key"))

        for dll, origins in tmp_dict.items():
            mechanisms.append({"name": "user_environment_script", "origins": list(origins), "entry": dll})

        # Windows Group Policy Scripts
        folder_path = os.path.join(fsw.forensicstore_path, "WindowsGroupPolicyScripts/")
        if os.path.exists(folder_path):
            for file in os.listdir(folder_path):
                file_path = os.path.join(folder_path, file)
                file_origin = get_file_origin(fsw, file_path)[0]

                # Group Policy Logon and Logoff Scripts
                if file.endswith("GroupPolicy_User_Scripts_scripts.ini"):
                    process_ini_values(tmp_dict, file_path, "Logon", file_origin)
                    process_ini_values(tmp_dict, file_path, "Logoff", file_origin)

                # Group Policy Startup and Shutdown Scripts
                elif file.endswith("GroupPolicy_Machine_Scripts_scripts.ini"):
                    process_ini_values(tmp_dict, file_path, "Startup", file_origin)
                    process_ini_values(tmp_dict, file_path, "Shutdown", file_origin)

                # Filesystem Startup and Shutdown Scripts
                elif file.endswith(".exe"):
                    if "Shutdown" in file:
                        mechanisms.append({"name": "logon", "origins": [file_origin],
                                           "entry": file.partition("Shutdown_")[2]})
                    elif "Startup" in file:
                        mechanisms.append({"name": "logon", "origins": [file_origin],
                                           "entry": file.partition("Startup_")[2]})
                    elif "Logon" in file:
                        mechanisms.append({"name": "logon", "origins": [file_origin],
                                           "entry": file.partition("Logon_")[2]})
                    elif "Logoff" in file:
                        mechanisms.append({"name": "logon", "origins": [file_origin],
                                           "entry": file.partition("Logoff_")[2]})

        # Windows Startup Scripts
        tmp_vals = {}
        for artifact in fsw.get_artifacts("WindowsStartupScript"):
            for value in artifact.get("values", []):
                if value.get("name") == "Parameters":
                    tmp_vals.setdefault(artifact.get("key"), {}).update({"param": value.get("data")})
                elif value.get("name") == "Script":
                    tmp_vals.setdefault(artifact.get("key"), {}).update({"script": value.get("data")})
                else:
                    tmp_vals.setdefault(artifact.get("key"), {}).update({"script": value.get("data"), "param": None})

        for key, entries in tmp_vals.items():
            tmp_dict.setdefault(" ".join(filter(None, [entries.get("script"), entries.get("param")])), set()).add(key)

        # Windows Startup Folders
        for file in fsw.get_files("WindowsStartupFolders"):
            if file.get("name", "").endswith(".exe", None):
                mechanisms.append(
                    {"name": "logon", "origins": [file.get("origin").get("path")], "entry": file.get("name")})

        return [{"name": "logon", "origins": list(origins), "entry": entry}
                for entry, origins in tmp_dict.items()]



import os
from glob import glob

from discoruns.mechanism import Mechanism
from discoruns.wrapper.fs_wrapper import ForensicStoreWrapper


def collect(fstore_full_path: str) -> list:
    fsw = ForensicStoreWrapper(fstore_full_path)
    plugins_to_run = ['app_cert_dlls',
                      'app_init_dlls',
                      'application_shimming',
                      'authentication_packages',
                      'browser_extensions',
                      'codecs',
                      'com_hijacking',
                      'file_associations',
                      'image_hijacks',
                      'known_dlls',
                      'logon',
                      'netsh_helper',
                      'office_application_startup',
                      'port_monitors',
                      'powershell_profiles',
                      'run_keys',
                      'scheduled_tasks',
                      'screensaver',
                      'security_support_provider',
                      'services_and_drivers',
                      'sip_and_trust_provider_hijacking',
                      'time_providers',
                      'winlogon_helper',
                      'winsock_providers',
                      ]

    # Instantiate all "Mechanism" subclasses (plugins) and collect mechanisms
    mechanisms_folder = os.path.join(os.path.dirname(os.path.abspath(__file__)), "mechanisms")
    for file in glob(os.path.join(mechanisms_folder, "*py")):
        name = os.path.splitext(os.path.basename(file))[0]
        if name in plugins_to_run:
            __import__(f"discoruns.mechanisms.{name}")
    collected_mechanisms = []
    for mechanisms in Mechanism.__subclasses__():
        collected_mechanisms.extend(mechanisms().collect_mechanism(fsw))
    fsw.close()
    return collected_mechanisms



from discoruns.mechanism import Mechanism


class AuthenticationPackages(Mechanism):

    def collect_mechanism(self, fsw):
        tmp_reg_keys_dict = {}
        found_auth_packages = set()

        for artifact in fsw.get_artifacts("WindowsLSAAuthenticationPackages"):
            raw_artifact_values = [value.get("data") for value in artifact.get("values", [])
                                   if value.get("name") == "Authentication Packages" and value.get("data")]

            # Some value cleaning since it's boxed in something like this "['exec_1', 'exec_2']"
            for raw_artifact_value in raw_artifact_values:
                cleaned_value = raw_artifact_value.strip('[').strip(']').replace("\'", "").split(",")
                cleaned_value = set(elem.strip(" ") for elem in cleaned_value)

                # Avoiding duplicate entries
                diff = cleaned_value.difference(found_auth_packages)
                for elem in diff:
                    tmp_reg_keys_dict.setdefault(elem, []).append(artifact.get("key"))
                found_auth_packages.update(cleaned_value)

        return [{"name": "authentication_packages", "origins": origins, "entry": package}
                for package, origins in tmp_reg_keys_dict.items()]



from discoruns.mechanism import Mechanism


class AppInitDLLs(Mechanism):

    def collect_mechanism(self, fsw):
        tmp_dict = {}

        for artifact in fsw.get_artifacts("WindowsAppInitDLLs"):
            for value in artifact.get("values", []):
                if value.get("name", "").casefold() == 'AppInit_DLLs'.casefold() and value.get("data"):
                    tmp_dict.setdefault(value.get("data"), set()).add(artifact.get("key"))

        return [{"name": "app_init_dlls", "origins": list(origins), "entry": dll} for dll, origins in tmp_dict.items()]



import os

from discoruns.mechanism import Mechanism


class ApplicationShimming(Mechanism):

    def collect_mechanism(self, fsw):
        tmp_dict = {}
        folder_path = os.path.join(fsw.forensicstore_path, "WindowsApplicationCompatibilityInstalledShimDatabases/")

        if os.path.exists(folder_path):
            for file in os.listdir(folder_path):
                if file.endswith(".sdb"):
                    for entry in fsw.get_files("WindowsApplicationCompatibilityInstalledShimDatabases"):
                        tmp_dict.setdefault(entry.get("name"), set()).add(entry.get("origin").get("path"))

        return [{"name": "application_shimming", "origins": list(path),
                 "entry": entry} for entry, path in tmp_dict.items()]



from discoruns.mechanism import Mechanism
from discoruns.wrapper.fs_wrapper import ForensicStoreWrapper


class SecuritySupportProvider(Mechanism):

    def collect_mechanism(self, fsw: ForensicStoreWrapper) -> list:
        tmp_dict = {}

        for artifact in fsw.get_artifacts("WindowsLSASecurityPackages"):
            for value in artifact.get("values", []):
                if value.get("data") and not value.get("data") == "['\"\"']":
                    tmp_dict.setdefault(value.get("data"), set()).add(artifact.get("key"))

        return [{"name": "security_support_provider", "origins": list(origins), "entry": dll}
                for dll, origins in tmp_dict.items()]



import argparse
import json
import logging
import os
import sys
import time
from pathlib import Path

import forensicstore
import tabulate

from discoruns import collect
from discoruns.wrapper.ac_wrapper import ArtifactCollectorWrapper

logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(
        prog="discoruns",
        description="Extract persistence mechanisms from disc images or forensicstores."
    )

    parser.add_argument(
        '--format',
        help="Output format of the collected persistence mechanisms.",
        choices=['console', 'json'],
        dest='format',
        default="console"
    )
    parser.add_argument(
        "-v", "--verbose",
        help="Show logging messages.",
        dest='verbose',
        action="store_true",
        default=False
    )

    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand", required=True)

    parser_image = subparsers.add_parser('image')
    parser_image.add_argument("image", metavar="IMAGE", help="Location of a disk image or a forensicstore.")

    parser_forensicstore = subparsers.add_parser('forensicstore')
    parser_forensicstore.add_argument("forensicstore", metavar="FORENSICSTORE", help="Location of a forensicstore.")

    args = parser.parse_args()

    if args.verbose:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    start_time = time.time()

    if args.subcommand == "image":
        if not os.path.exists(args.image):
            logger.error("Image does not exist!")
            sys.exit(1)

        fstore_full_path = f"{Path(args.image).stem}.forensicstore"

        # Create new store
        try:
            store = forensicstore.new(fstore_full_path)
            store.close()
        except forensicstore.forensicstore.StoreExitsError:
            logger.error("Forensicstore does allready exist!")
            sys.exit(1)

        # Start extraction
        logger.info("Collecting artifacts. This might take a while...")
        ArtifactCollectorWrapper(args.image, fstore_full_path,
                                 f"{os.path.join(os.path.dirname(os.path.abspath(__file__)), 'artifacts')}",
                                 "WindowsPersistence").collect_artifacts()

    elif args.subcommand == "forensicstore":
        if not os.path.exists(args.forensicstore):
            logger.error("Forensicstore does not exist!")
            sys.exit(1)

        logger.info("Using an existing forensicstore.")
        fstore_full_path = args.forensicstore

    collected_mechanisms = collect(fstore_full_path)

    if args.format == "json":
        print(json.dumps(collected_mechanisms, indent=4))
    else:
        print(tabulate.tabulate(collected_mechanisms))
        print()
        print("Total found entries: " + str(len(collected_mechanisms)))
        print("Execution time: %.4s seconds" % (time.time() - start_time))


if __name__ == '__main__':
    main()



import logging
import os
import os.path
import subprocess
import sys

logger = logging.getLogger(__name__)


class ArtifactCollectorWrapper:
    """
    This is a wrapper class for the artifact collect command-line tool
    """

    def __init__(self, image_path: str, store_path: str, artifacts_dir: str, artifact_names: str):
        """
        Initializes class with arguments from the command-line

        :param image_path: Path to the input disk image
        :param store_path: Path to the forensicstore output
        :param artifacts_dir: Path to the directory of the artifact definition yamls
        :param artifact_names: List of to-be collected artifacts
        """

        # File and directory paths. Should be also accessible from outside,
        # for example when trying to get output results
        self.image_path = image_path
        self.store_path = store_path
        self.artifacts_dir = artifacts_dir
        self.artifact_names = artifact_names

        # Docker mount points and image name. Could be placed in config file
        self._docker_image_mount = "/elementary/input-dir"
        self._docker_store_mount = "/elementary/input.forensicstore"
        self._docker_artifacts_mount = "/elementary/artifacts-dir"
        self.docker_image_name = "forensicanalysis/elementary-import-image:v0.3.5"

    def collect_artifacts(self):
        """
        Wrapper function for docker run.

        Calls the docker run with custom parameters as subprocess.
        """
        docker_input_mapping = f"{os.path.abspath(os.path.dirname(self.image_path))}:{self._docker_image_mount}"
        docker_output_mapping = f"{os.path.abspath(self.store_path)}:{self._docker_store_mount}"
        docker_artifacts_mapping = f"{os.path.abspath(self.artifacts_dir)}:{self._docker_artifacts_mount}"

        # Docker run subprocess call. "--user" tag makes sure that the output file has same permissions as the host,
        # to be able to access it afterwards.
        subprocess_commands = ["docker", "run", "--rm",
                               "-v", docker_input_mapping,
                               "-v", docker_artifacts_mapping,
                               "-v", docker_output_mapping,
                               f"{self.docker_image_name}",
                               "--input-file", f"{os.path.basename(self.image_path)}",
                               "--artifacts-dir", self._docker_artifacts_mount,
                               "--dir", self._docker_store_mount,
                               "--artifact", f"{self.artifact_names}",
                               "-vvv", f"--input-dir {self._docker_image_mount}"]

        logger.debug("Trying to execute docker run subprocess call to collect artifacts ...")
        logger.debug(" ".join(subprocess_commands))

        try:
            # pylint: disable=subprocess-run-check
            res = subprocess.run(subprocess_commands, capture_output=True)

            logger.debug("Subprocess call returncode: %s", res.returncode)

            # Subprocess return code handling
            if res.returncode == 0:
                if res.stdout:
                    logger.debug(res.stdout.decode("UTF-8"))
                if res.stderr:
                    logger.warning(res.stderr.decode("UTF-8"))
                logger.info("Successfully created forensicstore at: %s", self.store_path)
            else:
                if res.stdout:
                    logger.error(res.stdout.decode("UTF-8"))
                if res.stderr:
                    logger.error(res.stderr.decode("UTF-8"))
                sys.exit(1)

        except FileNotFoundError:
            logger.error("Could not run subprocess call. Maybe docker is not installed?")



from setuptools import setup

setup(
    name='discoruns',
    version='0.1.0',
    url='https://github.com/korbster/discoruns',
    author='Korbinian Karl',
    description='Python library for windows persistence mechanism extraction',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=['discoruns', "discoruns.artifacts", "discoruns.mechanisms", "discoruns.wrapper"],
    install_requires=[
        'forensicstore>=0.17.0,<0.18.0',
        'tabulate==0.8.7',
    ],
    include_package_data=True,
    entry_points={
        'console_scripts':
            ['discoruns = discoruns.main:main']
    },
    zip_safe=False,
)



from xml.dom import minidom

from discoruns.mechanism import Mechanism
from discoruns.wrapper.fs_wrapper import ForensicStoreWrapper


class ScheduledTasks(Mechanism):

    def collect_mechanism(self, fsw: ForensicStoreWrapper) -> list:
        mechanisms = []
        for artifact in fsw.get_files("WindowsScheduledTasks"):
            if "export_path" in artifact and artifact["export_path"] and not artifact["export_path"].endswith(".DAT"):
                # pylint: disable=protected-access
                with fsw._fs_store.load_file(artifact["export_path"]) as io:
                    xml_file = minidom.parse(io)
                    elem = ""
                    arg = ""

                    # Check xml doc for "Command" and "Arguments" tags. Exception wil be ignored if no argument exists.
                    try:
                        elem = xml_file.getElementsByTagName('Command')[0].childNodes[0].data
                        arg = xml_file.getElementsByTagName('Arguments')[0].childNodes[0].data
                    except IndexError:
                        pass

                    if elem:
                        mechanisms.append({
                            "name": "scheduled_tasks",
                            "origins": [artifact.get("origin").get("path")],
                            "entry": " ".join([elem, arg])}
                        )

        return mechanisms
